#-------------------------------------------------------------------
# tarfile.py
#
# Module for reading and writing .tar and tar.gz files.
#
# Needs at least Python version 2.0.
# NOTE: For Python versions 2.0 and 2.1, I added gzip22.py
# to the tarfile distribution, which is Python 2.2's gzip
# standard module. It adds support for seek() on
# GzipFile objects, which is absolutely essential.
#
# On the whole, tarfile.py is compatible to the zipfile
# standard module.
#
# Please consult the html documentation in this distribution
# for further details on how to use tarfile.
#
# Copyright (C) 2002 Lars Gustbel <lars@gustaebel.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#
#-------------------------------------------------------------------
# History:
# - 0.2.6   added some patches from Detlef Lannert, fixed some bugs
# - 0.2.5   added the stdout class for gzipped writing to sys.stdout.
#           kicked out the last bugs.
#           first public release.
# - 0.2.4   can now handle GNUtar longlinks.
# - 0.2.3   some cosmetic code surgery.
#           set up the documentation.
# - 0.2.2   can now handle GNUtar longnames.
#           still needs some testing.
# - 0.2     many many changes.
# - 0.1     initial version.
#
#-------------------------------------------------------------------
# Limitations:
# - appending to a gzipped tarfile is not possible.
#   (which is quite logical)
#
# - (GNU)Tar extras like sparse files, multiple volumes
#   etc. are not supported.
#
#-------------------------------------------------------------------

"""read from and write to (gzipped) TAR archives
"""

__author__ = "Lars Gustbel <lars@gustaebel.de>"
__date__ = "$Date: 2002/03/15 20:58:18 $"
__version__ = "0.2.6"
__cvsid__ = "$Id: tarfile.py,v 1.10 2002/03/15 20:58:18 lars Exp $"
__credits__ = "Detlef Lannert for some early patches"

import sys
import os
import os.path
import shutil
import stat
import time
from calendar import timegm

try:
    import grp
    import pwd
except ImportError:
    grp = None
    pwd = None

# For gzip compression we need at least the gzip module
# which comes with Python 2.2.
# To remain backward compatible (2.0, 2.1), I put
# gzip22.py in the tarfile distribution, which is
# Python 2.2's gzip.py. It is imported on demand.
if sys.version_info[1] < 2:                         # Python version < 2.2?
    try: import gzip22 as gzipmod                   # -> import the gzip22.py,
    except ImportError: gzipmod = None              # else disable gzipping.
else:
    import gzip as gzipmod                          # Python >= 2.2
                                                    # import the available gzip module

__all__ = ["TarFile", "TarInfo", "is_tarfile", "gzip", "gunzip", "error"]

#------------------------
# TAR specific constants
#------------------------
BLOCKSIZE = 512                 # length of processing blocks
RECORDSIZE = BLOCKSIZE * 20     # length of records (a product of BLOCKSIZE)
MAGIC = "ustar"                 # magic tar string
VERSION = "00"                  # version number

LENGTH_NAME = 100               # maximal length of a filename
LENGTH_LINK = 100               # maximal length of a linkname

REGTYPE = "0"                   # regular file
AREGTYPE = "\0"                 # regular file
LNKTYPE = "1"                   # link (inside tarfile)
SYMTYPE = "2"                   # symbolic link
CHRTYPE = "3"                   # character special device
BLKTYPE = "4"                   # block special device
DIRTYPE = "5"                   # directory
FIFOTYPE = "6"                  # fifo special device
CONTTYPE = "7"                  # contiguous file

GNUTYPE_LONGNAME = "L"          # GNUtar longnames
GNUTYPE_LONGLINK = "K"          # GNUtar longlink

#-------------------------------
# tarfile.py specific constants
#-------------------------------
# in compatibility to zipfile module
# these constants are used for __init__() of class TarFile
TAR_PLAIN = ZIP_STORED = 0          # plain tar file   (.tar)
TAR_GZIPPED = ZIP_DEFLATED = 8      # gzipped tar file (.tar.gz)

SUPPORTED_TYPES = (REGTYPE, AREGTYPE, LNKTYPE,  # file types that tarfile module
                   SYMTYPE, DIRTYPE, FIFOTYPE,  # can cope with.
                   CONTTYPE, GNUTYPE_LONGNAME,
                   GNUTYPE_LONGLINK)

#-----------------------
# Some useful functions
#-----------------------
def nts(s):
    """Convert a null-terminated string buffer to
    a normal python string."""
    return s.split("\0", 1)[0]

def stn(s, maxlen = None):
    """Reverse method to nts. Pad with zero-chars up to
    maxlen (optional)."""
    if not maxlen: return s + "\0"
    return s + "\0" * (maxlen - len(s))

def calc_chksum(buf):
    """Calculate the checksum for an entry's header. It's a simple addition
    of all bytes, treating the chksum field as if filled with spaces.
    buf is a 512 byte long string buffer which holds the header.
    """
    chk = 256                           # chksum field is treated as blanks,
                                        # so the initial value is 8 * ord(" ")
    for c in buf[:148]: chk += ord(c)   # sum up all bytes before chksum
    for c in buf[156:]: chk += ord(c)   # sum up all bytes after chksum
    return chk

def set_binary():
    """Win32 specific function. Sets stdout to binary.
    Use this function if you want to write a gzipped tarfile to
    sys.stdout on windows platforms. Call it before
    any output has been done. It is platform independent.
    """
    try:
        import msvcrt
        msvcrt.setmode(1, os.O_BINARY)
    except: pass

def test_file(file, mode):
    """Test if `file' has at least a read()/write() and seek() method.
    Return "r" for a read file, "w" for a write file.
    """
    if not hasattr(file, "seek"):
        return None
    if hasattr(file, "read") and mode[0:1] == "r":
        return "r"
    if hasattr(file, "write") and mode[0:1] == "w":
        return "w"
    return None

def copyfileobj(fo1, fo2, length):
    """Copy length bytes from fileobj fo1 to fileobj fo2.
    """
    BUFSIZE = 16384
    blocks, remainder = divmod(length, BUFSIZE)
    for b in range(blocks):
        buf = fo1.read(BUFSIZE)
        if buf is None: raise IOError, "end of file (EOF) reached"
        fo2.write(buf)

    buf = fo1.read(remainder)
    if buf is None: raise IOError, "end of file (EOF) reached"
    fo2.write(buf)    

if os.sep != "/":
    def normpath(path):
        return os.path.normpath(path).replace(os.sep, "/")
else:
    normpath = os.path.normpath

#--------------------     
# tarfile Exceptions
#--------------------
class BadTarFile(Exception):
    """Exception for bad tarfiles.
    """
    pass
error = (BadTarFile)

#--------------------
# exported functions
#--------------------
def is_tarfile(file):
    """Return true if file seems to be a valid tar archive.
    Empty files are treated like tar files. file should be either a
    filename or a file object.
    """
    close = 0
    if type(file) is type(""):
        if os.path.getsize(file) == 0: return 1
        close = 1
        file = open(file, "rb")
        if file.read(2) == "\037\213":
            file.seek(0)
            file = gzipmod.GzipFile(None, "rb", 9, file)
        else:
            file.seek(0)
    elif not test_file(file, "r"):
        raise IOError, "need a valid file or file-like object as argument"

    try: buf = file.read(BLOCKSIZE)
    except: return None

    if buf == "": return 1
    if buf[257:262] == MAGIC: return 1
    if close: file.close()
    return None

def gzip(file, gzfile = None):
    """gZip an existing file defined by its filename.
    """
    if not gzfile: gzfile = file + ".gz"
    gzfile = gzipmod.open(gzfile, "wb", 9)
    file = open(file, "rb", RECORDSIZE)
    while 1:
        buf = file.read(RECORDSIZE)
        if not buf: break
        gzfile.write(buf)
    gzfile.close()
    file.close()
    return

def gunzip(gzfile, file = None):
    """gUnzip a file defined by its filename.
    """
    if not file:
        file, ext = os.path.splitext(gzfile)
        if ext == ".tgz":
            file += ".tar"
        if ext != ".gz":
            file += ext      
    gzfile = gzipmod.open(gzfile, "rb")
    file = open(file, "wb", RECORDSIZE)
    while 1:
        buf = gzfile.read(RECORDSIZE)
        if not buf: break
        file.write(buf)
    file.close()
    gzfile.close()
    return


#------------------
# Exported Classes
#------------------
class TarInfo:
    """Informational class which represents a tar Header Block
    (POSIX 1003.1-1990). TarInfo is compatible to zipfile.ZipInfo
    with some variations and limitations. TarInfo objects are returned
    by TarFile.getinfo() and TarFile.infolist() and are normally created
    internally.
    If you want to create a TarInfo object from the outside, you
    use it similar to zipfile's ZipInfo.

    For object members, see below.
    """
    filename = None         # entry name (dirnames must end with '/')
    mode = 0100666          # file permissions (default)
    uid = 0                 # user id
    gid = 0                 # group id
    file_size = 0           # file size
    date_time = (0,)        # modification time as a tuple
    chksum = 0              # header checksum
    type = "0"              # entry type, default REGTYPE
    linkname = ""           # link name
    uname = ""              # user name
    gname = ""              # group name
    devmajor = None         #-+
    devminor = None         # + not neccessary
    prefix = None           #-+
    
    offset = 0              # the tar header starts here
    offset_data = 0         # the REGTYPE file's data starts here
    
    def __init__(self, filename = "", date_time = (1970, 1, 1, 0, 0, 0)):
        self.filename = filename
        self.date_time = date_time
        
    def FileHeader(self):
        """Return a tar header as a 512 byte string.
        """
        # The following code was provided by Detlef Lannert. Danke, Detlef.
        parts = []
        for value, fieldsize in (
                (self.filename, 100),                   # file name
                ("%07o" % self.mode, 8),                # file permissions
                ("%07o" % self.uid, 8),                 # user id
                ("%07o" % self.gid, 8),                 # group id
                ("%011o" % self.file_size, 12),         # file size
                ("%011o" % timegm(self.date_time), 12), # modification time
                ("        ", 8),                        # for checksum
                (self.type, 1),                         # file type
                (self.linkname, 100),                   # link name
                (MAGIC, 6),                             # magic string
                (VERSION, 2),                           # version number
                (self.uname, 32),                       # user name
                (self.gname, 32),                       # group name
                ):
            l = len(value)
            parts.append(value + (fieldsize - l) * "\0")
        
        buf = "".join(parts)
        chksum = calc_chksum(buf)
        buf = buf[:148] + "%07o" % chksum + buf[155:]
        buf += (512 - len(buf)) * "\0"
        self.buf = buf
        return buf
# class TarInfo


class TarFile:
    """Class representing a TAR archive on disk.
    """    
    debug = 0                   # may be set from 0 (no msgs) to 2 (all msgs)
    
    followsymlinks = 1          # if true, add content of linked file to tarfile,
                                # else the link. Is only used when writing.
    fp = None
    
    def __init__(self, file, mode = "r", compression = TAR_PLAIN):
        """Create a TarFile object.
        file must be a filename or a file(-like) object.
        mode must be one of `r', `a' or `w'.
        compression must be one of TAR_PLAIN or TAR_GZIPPED.
        """
        self.mode = mode[0:1]

        if compression not in (TAR_PLAIN, TAR_GZIPPED):
            raise RuntimeError, "Compression type is not supported"
        if compression == TAR_GZIPPED and not gzipmod:
            raise RuntimeError, "Accessing compressed TarFiles needs the gzip module"
        if self.mode not in ("raw"):
            raise RuntimeError, "Access mode must be one of `r', `a' or `w'"
        if type(file) is type(""):
            self.filename = file
            self.fp = self._open(file, self.mode, compression)
        elif test_file(file, self.mode):
            if hasattr(file, "name"):
                self.filename = file.name
            else:
                self.filename = ""
            self.fp = self._open(file, self.mode, compression)
        else:
            raise RuntimeError, "Need either a filename or a file object as first argument"

        if self.mode in "ra" and not is_tarfile(self.fp):
            raise BadTarFile, "File is not a tar file."

        # Init datastructures
        self.entries = []           # list of entires as TarInfo objects
        self.entriesnames = []      # names of entries 
        self.chunks = [0l]          # chunk cache
        self._loaded = 0            # indicates whether tarfile content has been fully loaded
        self.offset = 0l            # current position in the tarfile


    def __repr__(self):
        s = repr(self.fp)
        return '<tarfile ' + s[1:-1] + ' ' + hex(id(self)) + '>'        

    def _open(self, file, mode, compression):
        """Open a tarfile.
        """
        # Set the mode to binary explicitly
        mode = {"r": "rb", "a": "r+b", "w": "wb"}[mode]
        
        if compression == TAR_GZIPPED:
            if not gzipmod:
                raise RuntimeError, "gzip compression needs the gzip module"
            if mode == "r+b":
                raise RuntimeError, "Appending to gzipped TarFile is not possible"

            if type(file) is type(""):
                file = open(file, mode)
            if hasattr(file, "name"):
                filename = file.name
                filename, extension = os.path.splitext(filename)
                if extension == ".tgz": extension = ".tar"  # foo.tgz    -> foo.tar
                if extension == ".gz": extension = ""       # foo.tar.gz -> foo.tar
                filename += extension
            else:
                filename = ""
            return gzipmod.GzipFile(filename, mode, 9, file)
        else:
            if type(file) is type(""):
                return open(file, mode, RECORDSIZE)
            else:
                return file        
        
    def _load(self):
        """Read through the entire tarfile and look for valid entries.
        Is executed when e.g. infolist() or namelist() is called,
        when there is a random access via read() or extract().
        """
        while 1:
            tarinfo = self.next()
            if not tarinfo: break

        self._loaded = 1            
        self.fp.seek(0)
        return

    def _proc_gnulong(self, tarinfo, type):
        """Evaluate the two blocks that hold a longname.
        """
        buf = self.fp.read(BLOCKSIZE)
        if not buf: return None
        self.offset += BLOCKSIZE
        if type == GNUTYPE_LONGNAME:
            tarinfo.filename = nts(buf)
        if type == GNUTYPE_LONGLINK:
            tarinfo.linkname = nts(buf)

        buf = self.fp.read(BLOCKSIZE)
        if not buf: return None
        self.offset += BLOCKSIZE
        tarinfo.type = buf[156:157]
        return buf

    def next(self):
        """Return the next entry from the tarfile.
        Return None if the end of tarfile is reached.
        Can be used in a while statement and is used
        for Iteration (see __iter__()) and internally.
        """
        if not self.fp: raise IOError, "cannot access closed file"

        # Read the next block ignoring empty blocks.
        self.fp.seek(self.chunks[-1])
        while 1:
            buf = self.fp.read(BLOCKSIZE)
            if not buf: return None
            if buf[0] == "\0":
                self.offset += BLOCKSIZE
                continue
            break

        # We have got a tar header block.
        # Extract the information from it.
        tarinfo = TarInfo()
        tarinfo.filename = nts(buf[0:100])
        tarinfo.linkname = nts(buf[157:257])          
        tarinfo.type = buf[156:157]

        # If this header is a GNUTYPE longname or longlink
        # header, we must parse them first.
        if tarinfo.type == GNUTYPE_LONGLINK:
            buf = self._proc_gnulong(tarinfo, GNUTYPE_LONGLINK)
        if tarinfo.type == GNUTYPE_LONGNAME:
            buf = self._proc_gnulong(tarinfo, GNUTYPE_LONGNAME)
            
        tarinfo.mode = int(buf[100:108], 8)           
        tarinfo.uid = int(buf[108:116],8)             
        tarinfo.gid = int(buf[116:124],8)             
        tarinfo.file_size = long(buf[124:136], 8)          
        tarinfo.date_time = time.gmtime(long(buf[136:148], 8))[:6]    
        tarinfo.chksum = int(buf[148:156], 8)
        tarinfo.magic = buf[257:263]                  
        tarinfo.version = buf[263:265]                
        tarinfo.uname = nts(buf[265:297])             
        tarinfo.gname = nts(buf[297:329])             
        tarinfo.devmajor = buf[329:337]
        tarinfo.devminor = buf[337:345]
        tarinfo.prefix = buf[345:500]
        tarinfo.offset = self.offset
        tarinfo.offset_data = self.offset + BLOCKSIZE
        # The header is saved for later use of
        # testtar().
        tarinfo.buf = buf

        # Proceed with the next block.
        self.offset += BLOCKSIZE

        if tarinfo.type in (REGTYPE, AREGTYPE, CONTTYPE) and tarinfo.file_size:
            # If header contains a size field,
            # the following n data blocks are skipped.
            blocks, remainder = divmod(tarinfo.file_size, BLOCKSIZE)
            if remainder: blocks += 1
            self.offset += blocks * BLOCKSIZE

        self.entries.append(tarinfo)                # add the tarinfo object to members list
        self.entriesnames.append(tarinfo.filename)  # add the tarinfo's name to name list
        self.chunks.append(self.offset)             # add the position of the next possible entry
                                                    # to the chunk cache
        return tarinfo

    def _create_gnulong(self, filename, type):
        """Insert a GNU longname/longlink entry into a tarfile.
        It consists of a usual tar header, with the length
        of the filename as size, followed by data blocks (normally 1),
        which contain the filename as a null terminated string.
        """
        tarinfo = TarInfo()
        tarinfo.filename = "././@LongLink"
        tarinfo.type = type
        tarinfo.mode = 0
        self.writestr(tarinfo, filename + "\0")

    def infolist(self):
        """Return a list of all entries in tarfile (as TarInfo objects).
        """
        if not self._loaded: self._load()   # if we want to obtain a list of all members,
                                            # we have to read through the tarfile, first.
        return self.entries

    def namelist(self, force = 0):
        """Return a list of names of all entries in tarfile.
        """
        if not self._loaded: self._load()   # see comment in infolist()
        return self.entriesnames

    def __iter__(self):
        """Provide an iterator object for use in a for statement.
        Requires Python >= 2.2
        """
        if self._loaded:
            return self.entries.__iter__()  # all entries read, iterate over internal list
        else:
            return TarIter(self)            # return iterator

    def read(self, name):
        """Extract file `name' and return its bytes.
        """
        if self.mode == "w":
            raise IOError, "Cannot read from tarfile opened for writing"
        tarinfo = self.getinfo(name)
        self.fp.seek(tarinfo.offset_data)
        return self.fp.read(tarinfo.file_size)

    def testtar(self):
        """Test if all entries in tarfile are in good order.
        (This is quite a dumb check. We can only check
        if the headers are correct, not the data...)
        """
        if self.mode == "w":
            raise IOError, "Cannot test a tarfile opened for writing."
        bad = ""
        for tarinfo in self.infolist():
            if self.debug > 1: print tarinfo.filename,
            if self.debug == 2: print
            if calc_chksum(tarinfo.buf) != tarinfo.chksum:
                if self.debug > 2: print "bad"
                if not bad: bad = tarinfo.filename
            else:
                if self.debug > 2: print "ok"
        if bad: return bad
        return None
    testzip = testtar

    def extract(self, name, path = ""):
        """Extract file `name' from the tarfile and write
        it to path (optional). Instead of a filename,
        a TarInfo object can be passed.
        """
        if not self.fp: raise IOError, "cannot access closed file"

        # Fetch the TarInfo object for the given name
        # and build the destination pathname, replacing
        # forward slashes to platform specific separators
        if isinstance(name, TarInfo):
            tarinfo = name
            name = tarinfo.name
        else:
            tarinfo = self.getinfo(name)
        targetpath = os.path.join(path, name)
        if targetpath[-1:] == "/": targetpath = targetpath[:-1]
        targetpath = os.path.normpath(targetpath)


        # Extract the entry.
        if self.debug > 1: print targetpath

        if tarinfo.type not in SUPPORTED_TYPES:
            if self.debug > 1: print "(skipped: unsupported type)"
            return

        # If the entry is a directory, we are so nice
        # and create all upper directories on the fly.
        # os.makedirs() does a good job for that.
        if tarinfo.type == DIRTYPE:
            try: os.makedirs(targetpath)
            except (IOError, OSError), e: pass

        # The entry is regular file, resp. a contiguous file
        # which we treat like a regular one.
        # At first, all upper directories are created,
        # then we seek to the offset in the tarfile, where the
        # data locates and write it to the destination file.
        if tarinfo.type in (REGTYPE, AREGTYPE, CONTTYPE):
            try:
                os.makedirs(os.path.dirname(targetpath))
            except (IOError, OSError), e:
                pass
            try:
                self.fp.seek(tarinfo.offset_data)
                target = open(targetpath, "wb")
                copyfileobj(self.fp, target, tarinfo.file_size)
                target.close()
            except (IOError, OSError), e:
                if self.debug > 1: print "(%s)" % e
                return

        # The entry is either a symbolic link or a hardlink
        # (a link inside the tarfile).
        # If we want to create a symlink and the platform supports
        # it, we do this. If the platform does not support symlinks,
        # or if it is a hardlink, we first look inside the archive,
        # if the file can be found there, second we look a around
        # the filesystem.
        if tarinfo.type in (SYMTYPE, LNKTYPE):
            linkpath = tarinfo.linkname
            linkpath = os.path.normpath(linkpath)

            if hasattr(os, "symlink") and tarinfo.type == SYMTYPE:
                try: os.symlink(linkpath, targetpath)           # Try to create a symlink
                except (IOError, OSError), e:
                    if self.debug > 1: print "->", linkpath, "(%s)" % e
            else:
                linkpath = os.path.join(os.path.dirname(tarinfo.filename), tarinfo.linkname)
                linkpath = os.path.normpath(linkpath)   # cut away "../.././" and so on
                linkpath = normpath(linkpath)           # convert backslashes to slashes (on win32)
                try:
                    self.extract(linkpath, path)
                except (IOError, OSError, KeyError), e:
                    linkpath = os.path.normpath(linkpath)
                    try:
                        shutil.copy2(linkpath, targetpath)
                    except (IOError, OSError), e:
                        if self.debug > 1: print "->", linkpath, "(%s)" % e
                        return
            if self.debug > 1: print "->", linkpath

        # The entry is a fifo, so if the platform supports
        # fifos, we create one.
        if tarinfo.type in (FIFOTYPE,):
            if hasattr(os, "mkfifo"):
                try:
                    os.mkfifo(targetpath)
                except (IOError, OSError), e:
                    if self.debug > 1: print "(fifo) '%s'" % e
                    return
            else:
                if self.debug > 1: print "(fifo, not supported by os)"
                return
            if self.debug > 1: print "(fifo)"

        # All possible files are now extracted.

        # Here we set the file's last modification time.
        # XXX If file is a directory, we do not set its mtime
        # because in most of the cases, it is reset to
        # current time when accessing files underneath.
        # Seems to make problems under win32 anyway.
        if tarinfo.type != DIRTYPE:
            mtime = timegm(tarinfo.date_time)
            try:
                os.utime(targetpath, (mtime, mtime))
            except (IOError, OSError), e:
                if self.debug > 1: print "utime '%s'" % e

        # Here we set the ownership for the file.
        # We must be logged in as root. So, we
        # look for the owner on our system, and
        # if found, give it to him, else it's ours.
        if pwd and os.getgid() == 0:
            try: g = grp.getgrnam(tarinfo.gname)[2]
            except KeyError:
                try: g = grp.getgrgid(tarinfo.gid)[2]
                except KeyError: g = os.getgid()
            try: u = pwd.getpwnam(tarinfo.uname)[2]
            except KeyError:
                try: u = pwd.getpwuid(tarinfo.uid)
                except KeyError: u = os.getuid()
            try:
                os.chown(targetpath, u, g)
            except (IOError, OSError), e:
                if self.debug > 1: print "chown '%s'" % e

        # Here we set the file's permissions.
        try:
            os.chmod(targetpath, tarinfo.mode)
        except (IOError, OSError), e:
            if self.debug > 1: print "chmod '%s'" % e

    
    def writestr(self, tarinfo, data = "", fileobj = None):
        """Add the string buffer `data' to the tarfile.
        File information is passed by the tarinfo object.
        fileobj is used internally, for less memory consuming copying.
        """
        # If fileobj is not given, we write `data' to the archive.
        # So, the filesize is the length of the data. If fileobj
        # is given, tarinfo.file_size must have already been set.
        if not fileobj: tarinfo.file_size = len(data)

        # Try to set the username and groupname        
        if pwd: tarinfo.uname = pwd.getpwuid(tarinfo.uid)[0]
        else: tarinfo.uname = "user"
        if grp: tarinfo.gname = grp.getgrgid(tarinfo.gid)[0]
        else: tarinfo.gname = "group"

        # First, write out the tar header.        
        header = tarinfo.FileHeader()
        self.fp.write(header)
        self.offset += BLOCKSIZE

        # If there's data to follow, we append it.        
        if data:
            self.fp.write(data)
            blocks, remainder = divmod(tarinfo.file_size, BLOCKSIZE)
            if remainder > 0:
                self.fp.write("\0" * (BLOCKSIZE - remainder))
        elif fileobj:
            copyfileobj(fileobj, self.fp, tarinfo.file_size)
            blocks, remainder = divmod(tarinfo.file_size, BLOCKSIZE)
            if remainder > 0:
                self.fp.write("\0" * (BLOCKSIZE - remainder))
            

    def write(self, filename, arcname = None, compresstype = None, recursive = 1):
        """Add a file or a directory to the tarfile.
        compresstype is a dummy to remain compatible to zipfile.
        Directory addition is recursive by default.
        """
        if self.mode == "r":
            raise IOError, "Cannot write to a file opened for reading"

        # If the first file is about to be appended, we load hte table
        # of contents. Then we have to cut the trailing RECORDSIZE padding
        # off the tarfile, and start writing at that position.
        # See _open() when something goes wrong.
        if self.mode == "a" and not self._loaded:
            self.fp.seek(0)
            self._load()
            self.fp.seek(self.chunks[-1])

        # Skip if somebody tries to archive the archive...
        if os.path.abspath(filename) == os.path.abspath(self.filename):
            if self.debug > 2: print filename, "(skipped)"
            return

        # Special case: The user wants to add the current working directory.        
        if filename == ".":
            if recursive:
                if arcname == ".": arcname = ""
                for f in os.listdir("."):
                    self.write(f, os.path.join(arcname, f))
            return

        if self.debug > 1: print filename        

        # Building the name of the entry in the archive.
        # Backward slashes are converted to forward slashes,
        # Absolute paths are turned to relative paths.
        if not arcname: arcname = filename 
        arcname = normpath(arcname)
        drv, arcname = os.path.splitdrive(arcname)
        while arcname[0:1] == "/": arcname = arcname[1:]

        # Now, we are about to fill the TarInfo object with
        # information specific for the file.
        tarinfo = TarInfo()

        # Use os.stat or os.lstat, depending on platform
        # and if symlinks shall be resolved.
        if hasattr(os, "lstat") and not self.followsymlinks:
            osstat = os.lstat
        else:
            osstat = os.stat
            
        statres = osstat(filename)
        linkname = ""
        save_arcname = arcname

        # Here we test, what file type it is.        
        stmd = statres[stat.ST_MODE]
        if stat.S_ISREG(stmd):                  # a regular file?
            type = REGTYPE
        elif stat.S_ISDIR(stmd):                # a directory?
            type = DIRTYPE
            if arcname[-1:] != "/": arcname += "/"
        elif stat.S_ISFIFO(stmd):               # a fifo?
            type = FIFOTYPE
        elif stat.S_ISLNK(stmd):                # a symlink?
            type = SYMTYPE
            linkname = os.readlink(filename)
        else:
            if self.debug > 1: print "'%s' has an unsupported type" % filename
            return

        # Now we must check if the strings for filename
        # and linkname fit into the posix header. (100 chars for each)
        # If not, we must create GNUtar specific extra headers.
        # If both filename and linkname are too long,
        # the longlink is first to be written out.
        if len(linkname) >= LENGTH_LINK - 1:
            self._create_gnulong(linkname, GNUTYPE_LONGLINK)
            linkname = linkname[:LENGTH_LINK -1]
        if len(arcname) >= LENGTH_NAME - 1:
            self._create_gnulong(arcname, GNUTYPE_LONGNAME)
            arcname = arcname[:LENGTH_NAME - 1]

        # Fill the TarInfo with all information we can get.
        #
        tarinfo.filename = arcname                          # set filename in tarfile
        tarinfo.mode = stmd                                 # set file permissions
        tarinfo.uid = statres[stat.ST_UID]                  # set user id
        tarinfo.gid = statres[stat.ST_GID]                  # set group id
        tarinfo.file_size = statres[stat.ST_SIZE]           # set filesize
        tarinfo.date_time = time.gmtime(statres[stat.ST_MTIME])[:6] # set modification time
        tarinfo.type = type                                 # set file type
        tarinfo.linkname = linkname                         # set linkname
        if pwd: tarinfo.uname = pwd.getpwuid(tarinfo.uid)[0]# set username
        else: tarinfo.uname = "user"
        if grp: tarinfo.gname = grp.getgrgid(tarinfo.gid)[0]# set group name
        else: tarinfo.gname = "group"

        # In the end, we append the tar header and data
        # to the archive. If the file is a directory
        # we recurse its subdirectories (default).
        if tarinfo.type == REGTYPE:
            addf = open(filename, "rb")        
            self.writestr(tarinfo, fileobj = addf)
            addf.close()

        if tarinfo.type in (SYMTYPE, FIFOTYPE):
            tarinfo.size = 0l
            self.writestr(tarinfo, "")

        if tarinfo.type == DIRTYPE:
            self.writestr(tarinfo, "")
            if recursive: 
                for f in os.listdir(filename):
                    self.write(os.path.join(filename, f), os.path.join(save_arcname, f))        

    def getinfo(self, name):
        """Return a TarInfo object for file `name'.
        """
        # First, let's look if file is already in our
        # table of contents. (We sometimes don't want
        # to load the entire tarfile, but go through
        # it entry by entry.)
        # If that fails, we load all entries in the tarfile
        # and look there. (for random access)
        if name in self.entriesnames:
            i = self.entriesnames.index(name)
            return self.entries[i]
        else:
            namelist = self.namelist()
            if name in namelist:
                return self.entries[namelist.index(name)]
            else:
                raise KeyError, "filename `%s' is not in tarfile" % name
    
    def printdir(self):
        """Print a formatted listing of tarfile's contents to stdout.
        """
        print "%-46s %19s %12s" % ("File Name", "Modified    ", "Size")
        for tarinfo in self.infolist():
            date = "%d-%02d-%02d %02d:%02d:%02d" % tarinfo.date_time
            print "%-46s %s %12d" % (tarinfo.filename, date, tarinfo.file_size)

    def close(self):
        """Close the TarFile Object and do some cleanup.
        """
        if self.fp:
            if self.mode in "aw":
                # fill up the end with zero-blocks
                # (like option -b20 for tar does)
                blocks, remainder = divmod(self.offset, RECORDSIZE)
                if remainder > 0:
                    self.fp.write("\0" * (RECORDSIZE - remainder))
                    
            self.fp.close()
            self.fp = None
# class TarFile

class TarIter:
    """Iterator Class. Enables the use of TarFile
    in a for statement. This needs Python >= 2.2.
    
    for tarinfo in TarFile(...):
        suite...
    """
    
    def __init__(self, tarfile):
        """Construct TarIter object.
        """
        self.tarfile = tarfile
    def __iter__(self):
        """Return iteration object, in this case self.
        """
        return self
    def next(self):
        """Return the next item using TarFile's next() method.
        When all entries have been read, set TarFile as _loaded.
        """
        tarinfo = self.tarfile.next()
        if not tarinfo:
            self.tarfile._loaded = 1
            raise StopIteration
        return tarinfo
# class TarIter

class stdout:
    """This is a workaround when you want to write
    a gzipped TarFile to stdout.
    For example, if you want to create a file
    called "sources.tar.gz" and return it on the fly
    from a web-service, use stdout like that:
    
    tarfile = TarFile(stdout("sources.tar"), "w", TAR_GZIPPED)

    This is due to the fact, that if you pass a file-like
    object to gzip.GzipFile, it determines its filename by
    its attribute 'name'. This 'name' is then packed into
    the gzip-file.

    Please note that TarFile.debug is set to 0. (!!!)    

    For win32 users: sys.stdout is set to binary implicitly.    
    """
    def __init__(self, filename):
        set_binary()                # win32 specific, but portable
        self.name = filename        # this shall be my name.
    def write(self, s):
        sys.stdout.write(s)
    def seek(self, o = None): pass
#class stdout

if __name__ == "__main__":
    # a "light-weight" implementation of GNUtar ;-)
    import getopt
    opts, args = getopt.getopt(sys.argv[1:], "tcrzxf:C:")
    
    file = None
    dir = ""
    comp = 0
    dbg = 3
    for o, a in opts:
        if o == "-t": mode = "l"        # list archive
        if o == "-c": mode = "w"        # write to archive
        if o == "-r": mode = "a"        # append to archive
        if o == "-x": mode = "r"        # extract from archive
        if o == "-f": file = a          # specify filename else use stdout
        if o == "-C": dir = a           # change to dir
        if o == "-z": comp = 1          # filter through gzip

    if not file:
        dbg = 0 # suppress debug messages
        file = stdout("sys.stdout.tar")
    
    if comp: compression = TAR_GZIPPED
    else: compression = TAR_PLAIN
        
    if mode == "l":
        tarfile = TarFile(file, "r", compression)
    else:
        tarfile = TarFile(file, mode, compression)

    tarfile.debug = dbg

    if mode == "r":
        for t in tarfile.infolist():
            tarfile.extract(t.filename, dir)
    elif mode == "l":
        tarfile.printdir()
    else:
        for a in args:
            tarfile.write(a)

    tarfile.close()                
        
    